@@ -519,13 +519,13 @@ struct link {
 
 struct graph_port {
 	const struct fc_descriptor *desc;
-	void *hndl;
+	void **hndl;
 	uint32_t port;
 };
 
 struct graph_hndl {
 	const struct fc_descriptor *desc;
-	void *hndl;
+	void **hndl;
 };
 
 struct graph {
@@ -577,6 +577,10 @@ struct impl {
 	struct graph graph;
 };
 
+static int graph_instantiate(struct graph *graph);
+static void graph_cleanup(struct graph *graph);
+
+
 static void capture_destroy(void *d)
 {
 	struct impl *impl = d;
@@ -585,20 +589,26 @@ static void capture_destroy(void *d)
 }
 
 static void capture_process(void *d)
+{
+	struct impl *impl = d;
+	pw_stream_trigger_process(impl->playback);
+}
+
+static void playback_process(void *d)
 {
 	struct impl *impl = d;
 	struct pw_buffer *in, *out;
 	struct graph *graph = &impl->graph;
-	uint32_t i, outsize = 0, n_hndl = graph->n_hndl;
+	uint32_t i, insize = 0, outsize = 0, n_hndl = graph->n_hndl;
 	int32_t stride = 0;
 	struct graph_port *port;
 	struct spa_data *bd;
 
 	if ((in = pw_stream_dequeue_buffer(impl->capture)) == NULL)
-		pw_log_debug("out of capture buffers: %m");
+		pw_log_debug("%p: out of capture buffers: %m", impl);
 
 	if ((out = pw_stream_dequeue_buffer(impl->playback)) == NULL)
-		pw_log_debug("out of playback buffers: %m");
+		pw_log_debug("%p: out of playback buffers: %m", impl);
 
 	if (in == NULL || out == NULL)
 		goto done;
@@ -614,12 +624,14 @@ static void capture_process(void *d)
 		port = i < graph->n_input ? &graph->input[i] : NULL;
 
 		if (port && port->desc)
-			port->desc->connect_port(port->hndl, port->port,
+			port->desc->connect_port(*port->hndl, port->port,
 				SPA_PTROFF(bd->data, offs, void));
 
-		outsize = i == 0 ? size : SPA_MIN(outsize, size);
+		insize = i == 0 ? size : SPA_MIN(insize, size);
 		stride = SPA_MAX(stride, bd->chunk->stride);
 	}
+	outsize = insize;
+
 	for (i = 0; i < out->buffer->n_datas; i++) {
 		bd = &out->buffer->datas[i];
 
@@ -628,7 +640,7 @@ static void capture_process(void *d)
 		port = i < graph->n_output ? &graph->output[i] : NULL;
 
 		if (port && port->desc)
-			port->desc->connect_port(port->hndl, port->port, bd->data);
+			port->desc->connect_port(*port->hndl, port->port, bd->data);
 		else
 			memset(bd->data, 0, outsize);
 
@@ -636,9 +648,13 @@ static void capture_process(void *d)
 		bd->chunk->size = outsize;
 		bd->chunk->stride = stride;
 	}
+
+	pw_log_trace_fp("%p: stride:%d in:%d out:%d requested:%"PRIu64" (%"PRIu64")", impl,
+			stride, insize, outsize, out->requested, out->requested * stride);
+
 	for (i = 0; i < n_hndl; i++) {
 		struct graph_hndl *hndl = &graph->hndl[i];
-		hndl->desc->run(hndl->hndl, outsize / sizeof(float));
+		hndl->desc->run(*hndl->hndl, outsize / sizeof(float));
 	}
 
 done:
@@ -646,8 +662,6 @@ done:
 		pw_stream_queue_buffer(impl->capture, in);
 	if (out != NULL)
 		pw_stream_queue_buffer(impl->playback, out);
-
-	pw_stream_trigger_process(impl->playback);
 }
 
 static float get_default(struct impl *impl, struct descriptor *desc, uint32_t p)
@@ -682,11 +696,21 @@ static struct port *find_port(struct node *node, const char *name, int descripto
 	str = strdupa(name);
 	col = strchr(str, ':');
 	if (col != NULL) {
+		struct node *find;
 		node_name = str;
 		port_name = col + 1;
 		*col = '\0';
-		node = find_node(node->graph, node_name);
-	} else {
+		find = find_node(node->graph, node_name);
+		if (find == NULL) {
+			/* it's possible that the : is part of the port name,
+			 * try again without splitting things up. */
+			*col = ':';
+			col = NULL;
+		} else {
+			node = find;
+		}
+	}
+	if (col == NULL) {
 		node_name = node->name;
 		port_name = str;
 	}
@@ -736,11 +760,12 @@ static struct spa_pod *get_prop_info(struct graph *graph, struct spa_pod_builder
 	struct fc_port *p = &d->ports[port->p];
 	float def, min, max;
 	char name[512];
+	uint32_t rate = impl->rate ? impl->rate : 48000;
 
 	if (p->hint & FC_HINT_SAMPLE_RATE) {
-		def = p->def * impl->rate;
-		min = p->min * impl->rate;
-		max = p->max * impl->rate;
+		def = p->def * rate;
+		min = p->min * rate;
+		max = p->max * rate;
 	} else {
 		def = p->def;
 		min = p->min;
@@ -838,7 +863,7 @@ static int set_control_value(struct node *node, const char *name, float *value)
 
 	port = find_port(node, name, FC_PORT_INPUT | FC_PORT_CONTROL);
 	if (port == NULL)
-		return 0;
+		return -ENOENT;
 
 	node = port->node;
 	desc = node->desc;
@@ -853,7 +878,7 @@ static int parse_params(struct graph *graph, const struct spa_pod *pod)
 {
 	struct spa_pod_parser prs;
 	struct spa_pod_frame f;
-	int changed = 0;
+	int res, changed = 0;
 	struct node *def_node;
 
 	def_node = spa_list_first(&graph->node_list, struct node, link);
@@ -886,7 +911,8 @@ static int parse_params(struct graph *graph, const struct spa_pod *pod)
 			struct spa_pod *pod;
 			spa_pod_parser_get_pod(&prs, &pod);
 		}
-		changed += set_control_value(def_node, name, val);
+		if ((res = set_control_value(def_node, name, val)) > 0)
+			changed += res;
 	}
 	return changed;
 }
@@ -897,13 +923,14 @@ static void graph_reset(struct graph *graph)
 	for (i = 0; i < graph->n_hndl; i++) {
 		struct graph_hndl *hndl = &graph->hndl[i];
 		const struct fc_descriptor *d = hndl->desc;
+		if (hndl->hndl == NULL || *hndl->hndl == NULL)
+			continue;
 		if (d->deactivate)
-			d->deactivate(hndl->hndl);
+			d->deactivate(*hndl->hndl);
 		if (d->activate)
-			d->activate(hndl->hndl);
+			d->activate(*hndl->hndl);
 	}
 }
-
 static void param_props_changed(struct impl *impl, const struct spa_pod *param)
 {
 	struct spa_pod_object *obj = (struct spa_pod_object *) param;
@@ -978,8 +1005,15 @@ static void param_changed(void *data, uint32_t id, const struct spa_pod *param)
 
 	switch (id) {
 	case SPA_PARAM_Format:
-		if (param == NULL)
-			graph_reset(graph);
+		if (param == NULL) {
+			graph_cleanup(graph);
+		} else {
+			struct spa_audio_info_raw info;
+			spa_zero(info);
+			spa_format_audio_raw_parse(param, &info);
+			impl->rate = info.rate;
+			graph_instantiate(graph);
+		}
 		break;
 	case SPA_PARAM_Props:
 		if (param != NULL)
@@ -1009,6 +1043,7 @@ static void playback_destroy(void *d)
 static const struct pw_stream_events out_stream_events = {
 	PW_VERSION_STREAM_EVENTS,
 	.destroy = playback_destroy,
+	.process = playback_process,
 	.state_changed = state_changed,
 	.param_changed = param_changed
 };
@@ -1338,15 +1373,17 @@ static int parse_control(struct node *node, struct spa_json *control)
 	while (spa_json_get_string(control, key, sizeof(key)) > 0) {
 		float fl;
 		const char *val;
-		int len;
+		int res, len;
 
 		if ((len = spa_json_next(control, &val)) < 0)
 			break;
 
-		if (spa_json_parse_float(val, len, &fl) <= 0)
+		if (spa_json_parse_float(val, len, &fl) <= 0) {
 			pw_log_warn("control '%s' expects a number, ignoring", key);
-		else
-			set_control_value(node, key, &fl);
+		}
+		else if ((res = set_control_value(node, key, &fl)) < 0) {
+			pw_log_warn("control '%s' can not be set: %s", key, spa_strerror(res));
+		}
 	}
 	return 0;
 }
@@ -1464,6 +1501,7 @@ static int load_node(struct graph *graph, struct spa_json *json)
 	bool have_config = false;
 	uint32_t i;
 	int res;
+	float *data;
 
 	while (spa_json_get_string(json, key, sizeof(key)) > 0) {
 		if (spa_streq("type", key)) {
@@ -1524,6 +1562,10 @@ static int load_node(struct graph *graph, struct spa_json *json)
 	node->control_port = calloc(desc->n_control, sizeof(struct port));
 	node->notify_port = calloc(desc->n_notify, sizeof(struct port));
 
+	pw_log_info("loaded n_input:%d n_output:%d n_control:%d n_notify:%d",
+			desc->n_input, desc->n_output,
+			desc->n_control, desc->n_notify);
+
 	for (i = 0; i < desc->n_input; i++) {
 		struct port *port = &node->input_port[i];
 		port->node = node;
@@ -1538,6 +1580,14 @@ static int load_node(struct graph *graph, struct spa_json *json)
 		port->idx = i;
 		port->external = SPA_ID_INVALID;
 		port->p = desc->output[i];
+		if ((data = port->audio_data[i]) == NULL) {
+			data = calloc(1, MAX_SAMPLES * sizeof(float));
+			if (data == NULL) {
+				pw_log_error("cannot create port data: %m");
+				return -errno;
+			}
+		}
+		port->audio_data[i] = data;
 		spa_list_init(&port->link_list);
 	}
 	for (i = 0; i < desc->n_control; i++) {
@@ -1568,21 +1618,31 @@ static int load_node(struct graph *graph, struct spa_json *json)
 	return 0;
 }
 
-static void node_free(struct node *node)
+static void node_cleanup(struct node *node)
 {
-	uint32_t i, j;
 	const struct fc_descriptor *d = node->desc->desc;
+	uint32_t i;
 
-	spa_list_remove(&node->link);
 	for (i = 0; i < node->n_hndl; i++) {
-		for (j = 0; j < node->desc->n_output; j++)
-			free(node->output_port[j].audio_data[i]);
 		if (node->hndl[i] == NULL)
 			continue;
 		if (d->deactivate)
 			d->deactivate(node->hndl[i]);
 		d->cleanup(node->hndl[i]);
+		node->hndl[i] = NULL;
 	}
+}
+
+static void node_free(struct node *node)
+{
+	uint32_t i, j;
+
+	spa_list_remove(&node->link);
+	for (i = 0; i < node->n_hndl; i++) {
+		for (j = 0; j < node->desc->n_output; j++)
+			free(node->output_port[j].audio_data[i]);
+	}
+	node_cleanup(node);
 	descriptor_unref(node->desc);
 	free(node->input_port);
 	free(node->output_port);
@@ -1591,6 +1651,78 @@ static void node_free(struct node *node)
 	free(node);
 }
 
+static void graph_cleanup(struct graph *graph)
+{
+	struct node *node;
+	spa_list_for_each(node, &graph->node_list, link)
+		node_cleanup(node);
+}
+
+static int graph_instantiate(struct graph *graph)
+{
+	struct impl *impl = graph->impl;
+	struct node *node;
+	struct port *port;
+	struct link *link;
+	struct descriptor *desc;
+	const struct fc_descriptor *d;
+	uint32_t i, j;
+	int res;
+
+	spa_list_for_each(node, &graph->node_list, link) {
+		float *sd = silence_data, *dd = discard_data;
+
+		node_cleanup(node);
+
+		desc = node->desc;
+		d = desc->desc;
+		if (d->flags & FC_DESCRIPTOR_SUPPORTS_NULL_DATA)
+			sd = dd = NULL;
+
+		for (i = 0; i < node->n_hndl; i++) {
+			pw_log_info("instantiate %s %d rate:%lu", d->name, i, impl->rate);
+			if ((node->hndl[i] = d->instantiate(d, impl->rate, i, node->config)) == NULL) {
+				pw_log_error("cannot create plugin instance: %m");
+				res = -errno;
+				goto error;
+			}
+			for (j = 0; j < desc->n_input; j++) {
+				port = &node->input_port[j];
+				d->connect_port(node->hndl[i], port->p, sd);
+
+				spa_list_for_each(link, &port->link_list, input_link) {
+					struct port *peer = link->output;
+					pw_log_info("connect input port %s[%d]:%s %p",
+							node->name, i, d->ports[port->p].name,
+							peer->audio_data[i]);
+					d->connect_port(node->hndl[i], port->p, peer->audio_data[i]);
+				}
+			}
+			for (j = 0; j < desc->n_output; j++) {
+				port = &node->output_port[j];
+				pw_log_info("connect output port %s[%d]:%s %p",
+						node->name, i, d->ports[port->p].name,
+						port->audio_data[i]);
+				d->connect_port(node->hndl[i], port->p, port->audio_data[i]);
+			}
+			for (j = 0; j < desc->n_control; j++) {
+				port = &node->control_port[j];
+				d->connect_port(node->hndl[i], port->p, &port->control_data);
+			}
+			for (j = 0; j < desc->n_notify; j++) {
+				port = &node->notify_port[j];
+				d->connect_port(node->hndl[i], port->p, &port->control_data);
+			}
+			if (d->activate)
+				d->activate(node->hndl[i]);
+		}
+	}
+	return 0;
+error:
+	graph_cleanup(graph);
+	return res;
+}
+
 static struct node *find_next_node(struct graph *graph)
 {
 	struct node *node;
@@ -1603,61 +1735,16 @@ static struct node *find_next_node(struct graph *graph)
 	return NULL;
 }
 
-static int setup_input_port(struct graph *graph, struct port *port)
-{
-	struct descriptor *desc = port->node->desc;
-	const struct fc_descriptor *d = desc->desc;
-	struct link *link;
-	uint32_t i, n_hndl = port->node->n_hndl;
-
-	spa_list_for_each(link, &port->link_list, input_link) {
-		struct port *peer = link->output;
-		for (i = 0; i < n_hndl; i++) {
-			pw_log_info("connect input port %s[%d]:%s %p",
-					port->node->name, i, d->ports[port->p].name,
-					peer->audio_data[i]);
-			d->connect_port(port->node->hndl[i], port->p, peer->audio_data[i]);
-		}
-	}
-	return 0;
-}
-
-static int setup_output_port(struct graph *graph, struct port *port)
-{
-	struct descriptor *desc = port->node->desc;
-	const struct fc_descriptor *d = desc->desc;
-	struct link *link;
-	uint32_t i, n_hndl = port->node->n_hndl;
-
-	spa_list_for_each(link, &port->link_list, output_link) {
-		for (i = 0; i < n_hndl; i++) {
-			float *data;
-			if ((data = port->audio_data[i]) == NULL) {
-				data = calloc(1, MAX_SAMPLES * sizeof(float));
-				if (data == NULL)
-					return -errno;
-			}
-			port->audio_data[i] = data;
-			pw_log_info("connect output port %s[%d]:%s %p",
-					port->node->name, i, d->ports[port->p].name,
-					port->audio_data[i]);
-			d->connect_port(port->node->hndl[i], port->p, data);
-		}
-		link->input->node->n_deps--;
-	}
-	return 0;
-}
-
 static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_json *outputs)
 {
 	struct impl *impl = graph->impl;
 	struct node *node, *first, *last;
 	struct port *port;
+	struct link *link;
 	struct graph_port *gp;
 	struct graph_hndl *gh;
 	uint32_t i, j, n_nodes, n_input, n_output, n_control, n_hndl = 0;
 	int res;
-	unsigned long p;
 	struct descriptor *desc;
 	const struct fc_descriptor *d;
 	char v[256];
@@ -1728,52 +1815,11 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 	n_control = 0;
 	n_nodes = 0;
 	spa_list_for_each(node, &graph->node_list, link) {
-		float *sd = silence_data, *dd = discard_data;
-
+		node->n_hndl = n_hndl;
 		desc = node->desc;
-		d = desc->desc;
-		if (d->flags & FC_DESCRIPTOR_SUPPORTS_NULL_DATA)
-			sd = dd = NULL;
-
-		for (i = 0; i < n_hndl; i++) {
-			pw_log_info("instantiate %s %d", d->name, i);
-			if ((node->hndl[i] = d->instantiate(d, &impl->rate, i, node->config)) == NULL) {
-				pw_log_error("cannot create plugin instance: %m");
-				res = -errno;
-				goto error;
-			}
-			node->n_hndl = i + 1;
-
-			for (j = 0; j < desc->n_input; j++) {
-				p = desc->input[j];
-				d->connect_port(node->hndl[i], p, sd);
-			}
-			for (j = 0; j < desc->n_output; j++) {
-				p = desc->output[j];
-				d->connect_port(node->hndl[i], p, dd);
-			}
-			for (j = 0; j < desc->n_control; j++) {
-				port = &node->control_port[j];
-				d->connect_port(node->hndl[i], port->p, &port->control_data);
-			}
-			for (j = 0; j < desc->n_notify; j++) {
-				port = &node->notify_port[j];
-				d->connect_port(node->hndl[i], port->p, &port->control_data);
-			}
-			if (d->activate)
-				d->activate(node->hndl[i]);
-		}
 		n_control += desc->n_control;
 		n_nodes++;
 	}
-	pw_log_info("suggested rate:%lu capture:%d playback:%d", impl->rate,
-			impl->capture_info.rate, impl->playback_info.rate);
-
-	if (impl->capture_info.rate == 0)
-		impl->capture_info.rate = impl->rate;
-	if (impl->playback_info.rate == 0)
-		impl->playback_info.rate = impl->rate;
-
 	graph->n_input = 0;
 	graph->input = calloc(n_input * n_hndl, sizeof(struct graph_port));
 	graph->n_output = 0;
@@ -1789,7 +1835,7 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 				pw_log_info("input port %s[%d]:%s",
 						first->name, i, d->ports[desc->input[j]].name);
 				gp->desc = d;
-				gp->hndl = first->hndl[i];
+				gp->hndl = &first->hndl[i];
 				gp->port = desc->input[j];
 			}
 		} else {
@@ -1823,7 +1869,7 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 							port->node->name, i, d->ports[port->p].name);
 					port->external = graph->n_input;
 					gp->desc = d;
-					gp->hndl = port->node->hndl[i];
+					gp->hndl = &port->node->hndl[i];
 					gp->port = port->p;
 				}
 				graph->n_input++;
@@ -1837,7 +1883,7 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 				pw_log_info("output port %s[%d]:%s",
 						last->name, i, d->ports[desc->output[j]].name);
 				gp->desc = d;
-				gp->hndl = last->hndl[i];
+				gp->hndl = &last->hndl[i];
 				gp->port = desc->output[j];
 			}
 		} else {
@@ -1871,7 +1917,7 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 							port->node->name, i, d->ports[port->p].name);
 					port->external = graph->n_output;
 					gp->desc = d;
-					gp->hndl = port->node->hndl[i];
+					gp->hndl = &port->node->hndl[i];
 					gp->port = port->p;
 				}
 				graph->n_output++;
@@ -1891,17 +1937,16 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 		desc = node->desc;
 		d = desc->desc;
 
-		for (i = 0; i < desc->n_input; i++)
-			setup_input_port(graph, &node->input_port[i]);
-
 		for (i = 0; i < n_hndl; i++) {
 			gh = &graph->hndl[graph->n_hndl++];
-			gh->hndl = node->hndl[i];
+			gh->hndl = &node->hndl[i];
 			gh->desc = d;
-		}
 
-		for (i = 0; i < desc->n_output; i++)
-			setup_output_port(graph, &node->output_port[i]);
+		}
+		for (i = 0; i < desc->n_output; i++) {
+			spa_list_for_each(link, &node->output_port[i].link_list, output_link)
+				link->input->node->n_deps--;
+		}
 
 		/* collect all control ports on the graph */
 		for (i = 0; i < desc->n_control; i++) {
@@ -1909,17 +1954,8 @@ static int setup_graph(struct graph *graph, struct spa_json *inputs, struct spa_
 			graph->n_control++;
 		}
 	}
-	return 0;
-
+	res = 0;
 error:
-	spa_list_for_each(node, &graph->node_list, link) {
-		for (i = 0; i < node->n_hndl; i++) {
-			if (node->hndl[i] != NULL)
-				node->desc->desc->cleanup(node->hndl[i]);
-			node->hndl[i] = NULL;
-		}
-		node->n_hndl = 0;
-	}
 	return res;
 }
 
@@ -2050,12 +2086,20 @@ static const struct pw_proxy_events core_proxy_events = {
 
 static void impl_destroy(struct impl *impl)
 {
+	/* disconnect both streams before destroying any of them */
+	if (impl->capture)
+		pw_stream_disconnect(impl->capture);
+	if (impl->playback)
+		pw_stream_disconnect(impl->playback);
+
 	if (impl->capture)
 		pw_stream_destroy(impl->capture);
 	if (impl->playback)
 		pw_stream_destroy(impl->playback);
+
 	if (impl->core && impl->do_disconnect)
 		pw_core_disconnect(impl->core);
+
 	pw_properties_free(impl->capture_props);
 	pw_properties_free(impl->playback_props);
 	graph_free(&impl->graph);
@@ -2165,7 +2209,6 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	impl->module = module;
 	impl->context = context;
 
-	impl->rate = 48000;
 	impl->graph.impl = impl;
 	spa_list_init(&impl->plugin_list);
 
@@ -2175,6 +2218,8 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 		pw_properties_setf(props, PW_KEY_NODE_LINK_GROUP, "filter-chain-%u-%u", pid, id);
 	if (pw_properties_get(props, PW_KEY_NODE_VIRTUAL) == NULL)
 		pw_properties_set(props, PW_KEY_NODE_VIRTUAL, "true");
+	if (pw_properties_get(props, "resample.prefill") == NULL)
+		pw_properties_set(props, "resample.prefill", "true");
 	if (pw_properties_get(props, PW_KEY_NODE_DESCRIPTION) == NULL)
 		pw_properties_setf(props, PW_KEY_NODE_DESCRIPTION, "filter-chain-%u-%u", pid, id);
 
@@ -2192,6 +2237,7 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	copy_props(impl, props, PW_KEY_NODE_LATENCY);
 	copy_props(impl, props, PW_KEY_NODE_VIRTUAL);
 	copy_props(impl, props, PW_KEY_MEDIA_NAME);
+	copy_props(impl, props, "resample.prefill");
 
 	parse_audio_info(impl->capture_props, &impl->capture_info);
 	parse_audio_info(impl->playback_props, &impl->playback_info);
